#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Used to load in all the carts environment variables.

Wrapped all in if statements so that they can be used in
unit test environment
"""
from os.path import expanduser, join
from os import getenv


CONFIG_FILE = getenv('CARTD_CONFIG', join(
    expanduser('~'), '.pacifica-cartd', 'config.ini'))
CHERRYPY_CONFIG = getenv('CARTD_CPCONFIG', join(
    expanduser('~'), '.pacifica-cartd', 'cpconfig.ini'))



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Module that has the utility functionality for the cart."""
from __future__ import absolute_import
import os
import json
import datetime
import errno
import tarfile
from math import floor
import shutil
import psutil
import six
from peewee import DoesNotExist
from .orm import Cart, File
from .config import get_config

# pylint: disable=invalid-name
int_type = six.integer_types[-1]
# pylint: enable=invalid-name


class Cartutils(object):
    """Class used to provide utility functions for the cart to use."""

    def __init__(self):
        """Default constructor setting environment variable defaults."""
        self._vol_path = get_config().get('cartd', 'volume_path')
        self._lru_buff = get_config().get('cartd', 'lru_buffer_time')

    ###########################################################################
    #
    # Helper methods for handling cart path creation
    #
    ###########################################################################
    @staticmethod
    def fix_absolute_path(filepath):
        """Remove / from front of path."""
        if os.path.isabs(filepath):
            filepath = filepath[1:]
        return filepath

    @staticmethod
    def create_bundle_directories(filepath):
        """Create all the directories in the given path if they do not already exist."""
        try:
            os.makedirs(filepath, 0o777)
        except OSError as exception:
            # dont worry about error if the directory already exists
            # other errors are a problem however so push them up
            if exception.errno != errno.EEXIST:
                raise exception

    @classmethod
    def create_download_path(cls, cart_file, mycart, abs_cart_file_path):
        """Create the directories that the file will be pulled to."""
        try:
            cart_file_dirs = os.path.dirname(abs_cart_file_path)
            cls.create_bundle_directories(cart_file_dirs)
        except OSError as ex:
            cart_file.status = 'error'
            cart_file.error = 'Failed directory create with error: ' + str(ex)
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return False

        return True

    ###########################################################################
    #
    # Helper methods that determine space available/size
    # needed for the cart and files
    #
    ###########################################################################

    @staticmethod
    def check_file_size_needed(response, cart_file, mycart):
        """Check response (should be from Archive Interface head request) for file size."""
        try:
            decoded = json.loads(response)
            filesize = decoded['filesize']
            return int_type(filesize)
        except (ValueError, KeyError, TypeError) as ex:
            cart_file.status = 'error'
            cart_file.error = """Failed to decode file size
            json with error: """ + str(ex) + """ Response received from the
            Archive is: """ + str(response)
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return -1

    def check_space_requirements(self, cart_file, mycart, size_needed, deleted_flag):
        """
        Check to make sure there is enough space available on disk for the file to be downloaded.

        Note it will recursively call itself if there isnt enough
        space. It will delete a cart first, then call  itself
        until either there is enough space or there is no carts to delete
        """
        try:
            # available space is in bytes
            available_space = int_type(psutil.disk_usage(self._vol_path).free)
        except psutil.Error as ex:
            cart_file.status = 'error'
            cart_file.error = """Failed to get available file
            space with error: """ + str(ex)
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return False

        if size_needed > available_space:
            if deleted_flag:
                cart_deleted = self.lru_cart_delete(mycart)
                return self.check_space_requirements(cart_file, mycart,
                                                     size_needed, cart_deleted)
            cart_file.status = 'error'
            cart_file.error = 'Not enough space to download file'
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return False

        # there is enough space so return true
        return True

    @classmethod
    def get_path_size(cls, source):
        """Return the size of a specific directory, including all subdirectories and files."""
        total_size = os.path.getsize(source)
        for item in os.listdir(source):
            itempath = os.path.join(source, item)
            if os.path.isfile(itempath):
                total_size += os.path.getsize(itempath)
            elif os.path.isdir(itempath):
                total_size += cls.get_path_size(itempath)
        return total_size

    ###########################################################################
    #
    # Helper methods that parse the Archive Interface Responses
    #
    ###########################################################################

    def check_file_ready_pull(self, response, cart_file, mycart):
        """
        Check file ready state.

        Check response (should be from Archive Interface head request)
        for bytes per level then returns True or False based on if the file
        is at level 1 (downloadable)
        """
        size_needed = self.check_file_size_needed(response, cart_file, mycart)
        mod_time = self.check_file_modified_time(response, cart_file, mycart)
        try:
            decoded = json.loads(response)
            media = decoded['file_storage_media']
            if media == 'disk':
                return self.check_status_details(mycart, cart_file, size_needed, mod_time)
            return False
        except (ValueError, KeyError, TypeError) as ex:
            cart_file.status = 'error'
            cart_file.error = """Failed to decode json for file status
            with error: """ + str(ex) + """ Response received from the
            Archive is: """ + str(response)
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return -1

    def check_status_details(self, mycart, cart_file, size_needed, mod_time):
        """
        Check to see if status response is correct.

        Data from the status response is all correct and
        ready to for the file to be pulled.
        """
        # Return from function if the values couldnt be parsed (-1 return)
        if size_needed < 0 or mod_time < 0:
            return -1
        # set up saving path and return dictionary
        abs_cart_file_path = os.path.join(
            self._vol_path, str(mycart.id), mycart.cart_uid, cart_file.bundle_path)
        path_created = self.create_download_path(
            cart_file, mycart, abs_cart_file_path)
        # Check size here and make sure enough space is available.
        enough_space = self.check_space_requirements(
            cart_file, mycart, size_needed, True)
        if path_created and enough_space:
            return {'modtime': mod_time, 'filepath': abs_cart_file_path,
                    'path_created': path_created, 'enough_space': enough_space}
        return -1

    @staticmethod
    def check_file_modified_time(response, cart_file, mycart):
        """
        Check response for file modified time.

        Should be from Archive Interface head request
        """
        try:
            decoded = json.loads(response)
            mod_time = floor(float(decoded['mtime']))
            return mod_time
        except (ValueError, KeyError, TypeError) as ex:
            cart_file.status = 'error'
            cart_file.error = """Failed to decode file mtime
            json with error: """ + str(ex) + """ Response received from the
            Archive is: """ + str(response)
            cart_file.save()
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
            return -1

    ###########################################################################
    #
    # Helper methods used to delete carts
    #
    ###########################################################################

    def remove_cart(self, uid):
        """
        Call when a DELETE request comes in.

        Verifies there is a cart to delete then removes it.
        """
        deleted_flag = True
        iterator = 0  # used to verify at least one cart deleted
        carts = (Cart
                 .select()
                 .where(
                     (Cart.cart_uid == str(uid)) &
                     (Cart.deleted_date.is_null(True))))
        for cart in carts:
            iterator += 1
            success = self.delete_cart_bundle(cart)
            if not success:
                deleted_flag = False
        if deleted_flag and iterator > 0:
            return 'Cart Deleted Successfully'
        elif deleted_flag:
            return False  # already deleted
        return None  # unknown error

    def delete_cart_bundle(self, cart):
        """
        Get the path to where a carts file are.

        Also attempt to delete the file tree.
        """
        try:
            path_to_files = os.path.join(self._vol_path, str(cart.id))
            shutil.rmtree(path_to_files)
            cart.status = 'deleted'
            cart.deleted_date = datetime.datetime.now()
            cart.save()
            return True
        except OSError:
            return False

    def lru_cart_delete(self, mycart):
        """
        Delete the least recently used cart that isnt this one.

        Only delete one cart per call.
        """
        try:
            lru_time = datetime.datetime.now() - datetime.timedelta(
                seconds=int(self._lru_buff))
            del_cart = (Cart
                        .select()
                        .where(
                            (Cart.id != mycart.id) &
                            (Cart.deleted_date.is_null(True)) &
                            (Cart.updated_date < lru_time))
                        .order_by(Cart.creation_date)
                        .get())
            return self.delete_cart_bundle(del_cart)
        except DoesNotExist:
            # case if no cart exists that can be deleted
            return False

    ###########################################################################
    #
    # Cart Interface helpers for returning status/download paths
    #
    ###########################################################################

    @staticmethod
    def cart_status(uid):
        """Get the status of a specified cart."""
        status = None
        try:
            mycart = (Cart
                      .select()
                      .where(
                          (Cart.cart_uid == str(uid)) &
                          (Cart.deleted_date.is_null(True)))
                      .order_by(Cart.creation_date.desc())
                      .get())
        except DoesNotExist:
            # case if no record exists yet in database
            mycart = None
            status = ['error', 'No cart with uid ' + uid + ' found']

        if mycart:
            # send the status and any available error text
            status = [mycart.status, mycart.error]

        return status

    @staticmethod
    def available_cart(uid):
        """
        Check if the asked for cart tar is available.

        Returns the path to tar if yes, false if not. None if no cart.
        """
        cart_bundle_path = False
        try:
            mycart = (Cart
                      .select()
                      .where(
                          (Cart.cart_uid == str(uid)) &
                          (Cart.deleted_date.is_null(True)))
                      .order_by(Cart.creation_date.desc())
                      .get())
        except DoesNotExist:
            # case if no record exists yet in database
            mycart = None
            cart_bundle_path = None

        if mycart and mycart.status == 'ready':
            cart_bundle_path = mycart.bundle_path
        return cart_bundle_path

    ###########################################################################
    #
    # Helpers that update a carts files and a cart/file error
    #
    ###########################################################################

    @staticmethod
    def set_file_status(cart_file, cart, status, error):
        """Set the status and/or error for a cart."""
        cart_file.status = str(status)
        if error:
            cart_file.error = str(error)
        cart_file.save()
        cart.updated_date = datetime.datetime.now()
        cart.save()

    @classmethod
    def update_cart_files(cls, cart, file_ids):
        """Update the files associated to a cart."""
        with Cart.atomic():
            for f_id in file_ids:
                try:
                    filepath = cls.fix_absolute_path(f_id['path'])
                    hashtype = f_id['hashtype']
                    hashval = f_id['hashsum']
                    File.create(cart=cart, file_name=f_id['id'], bundle_path=filepath,
                                hash_type=hashtype, hash_value=hashval)
                    cart.updated_date = datetime.datetime.now()
                    cart.save()
                except (NameError, KeyError) as ex:
                    return ex  # return error so that the cart can be updated
            return None

    def prepare_bundle(self, cartid):
        """
        Check to see if all the files are staged locally.

        Before calling the bundling action.  If not will call
        itself to continue the waiting process
        """
        bundle_flag = True
        for c_file in File.select().where(File.cart == cartid):
            if c_file.status == 'error':
                # error pulling file so set cart error and return
                try:
                    mycart = Cart.get(Cart.id == cartid)
                    mycart.status = 'error'
                    mycart.error = 'Failed to pull file({})'.format(
                        c_file.error)
                    mycart.updated_date = datetime.datetime.now()
                    mycart.save()
                    Cart.database_close()
                    return
                except DoesNotExist:  # pragma: no cover
                    # case if record no longer exists
                    # creating this case in unit testing requires deletion and creation
                    # occuring nearly simultaneously, as such cant create unit test
                    Cart.database_close()
                    return

            elif c_file.status != 'staged':
                bundle_flag = False

        if bundle_flag:
            self.tar_files(cartid)

    def tar_files(self, cartid):
        """
        Start to bundle all the files together.

        The option to do streaming download or not is
        based on a system configuration.
        """
        try:
            mycart = Cart.get(Cart.id == cartid)
            if mycart.bundle:
                bundle_path = os.path.join(
                    self._vol_path,
                    str(mycart.id),
                    mycart.cart_uid
                )
                bundle_tar = '{}.tar'.format(bundle_path)
                cart_tar = tarfile.open(
                    bundle_tar,
                    mode='w'
                )
                cart_tar.add(
                    bundle_path, arcname=os.path.basename(bundle_path))
                cart_tar.close()
                shutil.rmtree(bundle_path)
                bundle_path = bundle_tar
            else:
                bundle_path = os.path.join(
                    self._vol_path,
                    str(mycart.id),
                    mycart.cart_uid
                )
            mycart.status = 'ready'
            mycart.bundle_path = bundle_path
            mycart.updated_date = datetime.datetime.now()
            mycart.save()
        except DoesNotExist:
                # case if record no longer exists
            Cart.database_close()
            return



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""File used to unit test the pacifica_cart tasks."""
import os
import unittest
import datetime
import json
import mock
import requests
from pacifica.cartd.orm import Cart, File
from pacifica.cartd.tasks import stage_file_task, stage_files, status_file_task, pull_file
from pacifica.cartd.archive_requests import ArchiveRequests
from pacifica.cartd.utils import Cartutils
from pacifica.cartd.tasks import CART_APP
from cart_db_setup_test import cart_dbsetup_gen

CART_APP.conf.CELERY_ALWAYS_EAGER = True


class TestTasks(cart_dbsetup_gen(unittest.TestCase)):
    """Contains tests for tasks that dont need all services stood up."""

    @mock.patch.object(ArchiveRequests, 'stage_file')
    def test_bad_stage(self, mock_stage_file):
        """Test the bad stage of a archive file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_stage_file.side_effect = requests.exceptions.RequestException(
            mock.Mock(status=500),
            'Error'
        )
        file_id = test_file.id
        stage_file_task(file_id)
        cart_file = File.get(File.id == file_id)
        status = cart_file.status
        self.assertEqual(status, 'error')

    @mock.patch.object(ArchiveRequests, 'status_file')
    @mock.patch.object(ArchiveRequests, 'stage_file')
    def test_bad_status(self, mock_stage_file, mock_status_file):
        """Test the bad status of a archive file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_stage_file.return_value = True
        mock_status_file.side_effect = requests.exceptions.RequestException(
            mock.Mock(status=500),
            'Error'
        )
        file_id = test_file.id
        status_file_task(file_id)
        cart_file = File.get(File.id == file_id)
        status = cart_file.status
        self.assertEqual(status, 'error')

    @mock.patch.object(os, 'utime')
    @mock.patch.object(ArchiveRequests, 'pull_file')
    @mock.patch.object(ArchiveRequests, 'status_file')
    @mock.patch.object(ArchiveRequests, 'stage_file')
    def test_bad_pull(self, mock_stage_file, mock_status_file, mock_pull_file, mock_utime):
        """Test the bad pull of a archive file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_stage_file.return_value = True
        mock_status_file.return_value = json.dumps({
            'bytes_per_level': '(10L, 0L)',
            'ctime': '1444629567',
            'file': '1.txt',
            'file_storage_media': 'disk',
            'filesize': '10',
            'message': 'File was found',
            'mtime': '1444937154'
        })
        mock_pull_file.side_effect = requests.exceptions.RequestException(
            mock.Mock(status=500),
            'Error'
        )
        mock_utime.return_value = True
        file_id = test_file.id
        stage_file_task(file_id)
        cart_file = File.get(File.id == file_id)
        status = cart_file.status
        self.assertEqual(status, 'error')

    @mock.patch.object(Cartutils, 'check_file_size_needed')
    @mock.patch.object(ArchiveRequests, 'status_file')
    @mock.patch.object(ArchiveRequests, 'stage_file')
    def test_bad_size_needed(self, mock_stage_file, mock_status_file, mock_check_file):
        """Test a error in check_file_size_needed."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_stage_file.return_value = True
        mock_status_file.return_value = json.dumps({
            'bytes_per_level': '(10L, 0L)',
            'ctime': '1444629567',
            'file': '1.txt',
            'file_storage_media': 'disk',
            'filesize': '10',
            'message': 'File was found',
            'mtime': '1444937154'
        })
        mock_check_file.return_value = -1
        file_id = test_file.id
        stage_file_task(file_id)
        cart_file = File.get(File.id == file_id)
        status = cart_file.status
        self.assertEqual(status, 'staging')

    @mock.patch.object(Cartutils, 'check_file_ready_pull')
    @mock.patch.object(ArchiveRequests, 'status_file')
    @mock.patch.object(ArchiveRequests, 'stage_file')
    def test_bad_ready_to_pull(self, mock_stage_file, mock_status_file, mock_check_file):
        """Test a error return from a file not ready to pull."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_stage_file.return_value = True
        mock_status_file.return_value = json.dumps({
            'bytes_per_level': '(10L, 0L)',
            'ctime': '1444629567',
            'file': '1.txt',
            'file_storage_media': 'disk',
            'filesize': '10',
            'message': 'File was found',
            'mtime': '1444937154'
        })
        mock_check_file.return_value = -1
        file_id = test_file.id
        stage_file_task(file_id)
        cart_file = File.get(File.id == file_id)
        status = cart_file.status
        self.assertEqual(status, 'staging')

    @mock.patch.object(Cartutils, 'update_cart_files')
    def test_bad_file_ids(self, mock_update):
        """Test a error return from a file not ready to pull."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_update.return_value = 'I am a error'
        file_id = test_file.id
        stage_files(file_id, test_cart.id)
        cart_after = Cart.get(Cart.id == test_cart.id)
        status = cart_after.status
        self.assertEqual(status, 'error')

    def test_stage_bad_file_id(self):
        """Test a error return from a file not ready to pull."""
        file_id = 9999999
        stage_file_task(file_id)
        self.assertEqual(True, True)

    def test_stage_cart_deleted(self):
        """Test a error return from a file not ready to pull."""
        test_cart = Cart.create(
            cart_uid='1',
            status='staging',
            deleted_date='2017-05-03 00:00:00'
        )
        test_file = File.create(
            cart=test_cart,
            file_name='1.txt',
            bundle_path='/tmp/1/1.txt'
        )
        file_id = test_file.id
        stage_file_task(file_id)
        self.assertEqual(True, True)

    def test_status_cart_deleted(self):
        """Test a error return from a file not ready to pull."""
        test_cart = Cart.create(
            cart_uid='1',
            status='staging',
            deleted_date='2017-05-03 00:00:00'
        )
        test_file = File.create(
            cart=test_cart,
            file_name='1.txt',
            bundle_path='/tmp/1/1.txt'
        )
        file_id = test_file.id
        status_file_task(file_id)
        self.assertEqual(True, True)

    @mock.patch.object(ArchiveRequests, 'pull_file')
    def test_bad_pull_value(self, mock_pull):
        """Test a error return from a file not ready to pull."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        mock_pull.side_effect = ValueError('Error with hash pulling file')
        file_id = test_file.id
        pull_file(file_id, '/tmp/1/1.txt', '9999999', False)
        cart_after = Cart.get(Cart.id == test_cart.id)
        status = cart_after.status
        self.assertEqual(status, 'error')

    @mock.patch.object(ArchiveRequests, 'pull_file')
    def test_bad_deleted_cart(self, mock_pull):
        """Test a error return from a file not ready to pull."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        test_cart.deleted_date = datetime.datetime.now()
        test_cart.save()
        pull_file(test_file.id, '/tmp/1/1.txt', '9999999', False)
        mock_pull.assert_not_called()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""File used to unit test the pacifica_cart."""
import unittest
import os
import json
from types import MethodType
import shutil
import mock
import psutil
from pacifica.cartd.orm import Cart
from pacifica.cartd.utils import Cartutils
import pacifica.cartd.orm
from cart_db_setup_test import cart_dbsetup_gen


class TestUtils(cart_dbsetup_gen(unittest.TestCase)):
    """Contains all the tests for the CartUtils class."""

    def test_create_download_path(self):
        """Test the creation of the download path for a cart file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        success = cart_utils.create_download_path(
            test_file,
            test_cart,
            test_file.bundle_path
        )
        directory_name = os.path.dirname(test_file.bundle_path)
        self.assertEqual(success, True)
        self.assertEqual(os.path.isdir(directory_name), True)

        os.rmdir(directory_name)
        self.assertEqual(os.path.isdir(directory_name), False)

    @mock.patch.object(Cartutils, 'create_bundle_directories')
    def test_bad_create_download_path(self, mock_create_bundle):
        """Test the creation of the download path for a cart file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        mock_create_bundle.side_effect = OSError(mock.Mock(), 'Error')
        success = cart_utils.create_download_path(
            test_file,
            test_cart,
            test_file.bundle_path
        )
        self.assertEqual(success, False)

    def test_create_bundle_directories(self):
        """Test the  creation of directories where files will be saved."""
        directory_name = '/tmp/fakedir/'
        cart_utils = Cartutils()
        cart_utils.create_bundle_directories(directory_name)
        self.assertEqual(os.path.isdir(directory_name), True)
        os.rmdir(directory_name)
        self.assertEqual(os.path.isdir(directory_name), False)

    @mock.patch.object(os, 'makedirs')
    def test_bad_makedirs(self, mock_makedirs):
        """Test a error return from a file not ready to pull."""
        mock_makedirs.side_effect = OSError(mock.Mock(), 'Error')
        c_util = Cartutils()
        self.assertRaises(
            OSError,
            c_util.create_bundle_directories,
            'fakepath'
        )

    def test_fix_absolute_path(self):
        """Test the correct creation of paths by removing absolute paths."""
        cart_utils = Cartutils()
        return_one = cart_utils.fix_absolute_path('tmp/foo.text')
        return_two = cart_utils.fix_absolute_path('/tmp/foo.text')
        return_three = cart_utils.fix_absolute_path('/tmp/foo.text')
        self.assertEqual(return_one, 'tmp/foo.text')
        self.assertEqual(return_two, 'tmp/foo.text')
        self.assertNotEqual(return_three, '/tmp/foo.text')

    def test_check_file_size_needed(self):
        """Test that the file size returned from the archive is parsed right."""
        response = json.dumps({
            'bytes_per_level': '(24L, 0L, 0L, 0L, 0L)',
            'ctime': '1444938166',
            'file': '/myemsl-dev/bundle/file.1',
            'file_storage_media': 'disk',
            'filesize': '24',
            'message': 'File was found',
            'mtime': '1444938166'
        })
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        file_size = cart_utils.check_file_size_needed(
            response,
            test_file,
            test_cart
        )
        self.assertEqual(file_size, 24)
        self.assertNotEqual(test_file.status, 'error')

        # now check for an error by sending a bad response
        file_size = cart_utils.check_file_size_needed(
            '',
            test_file,
            test_cart
        )
        self.assertEqual(file_size, -1)
        self.assertEqual(test_file.status, 'error')

    def test_check_space_requirements(self):
        """Test that there is enough space on the volume for the file."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        rtn = cart_utils.check_space_requirements(
            test_file,
            test_cart,
            10,
            False
        )
        self.assertEqual(rtn, True)
        self.assertNotEqual(test_file.status, 'error')

        # now check for an error by sending a way to large size needed number
        rtn = cart_utils.check_space_requirements(
            test_file,
            test_cart,
            9999999999999999999999,
            True
        )
        self.assertEqual(rtn, False)
        self.assertEqual(test_file.status, 'error')

    @mock.patch.object(psutil, 'disk_usage')
    def test_check_space_bad_path(self, mock_disk_usage):
        """Test that the error when a bad path."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        mock_disk_usage.side_effect = psutil.Error(mock.Mock())
        rtn = cart_utils.check_space_requirements(
            test_file,
            test_cart,
            10,
            False
        )
        self.assertEqual(rtn, False)
        self.assertEqual(test_file.status, 'error')

    def test_get_path_size(self):
        """Test to see if the path size of a directory is returned."""
        cart_utils = Cartutils()
        path = os.path.dirname(os.path.realpath(__file__))
        rtn = cart_utils.get_path_size(path + '/../')
        self.assertNotEqual(rtn, 0)

    def test_check_file_not_ready_pull(self):
        """Test that checks to see if a file is not ready to pull by checking the archive response."""
        response = json.dumps({
            'bytes_per_level': '(0L, 24L, 0L, 0L, 0L)',
            'ctime': '1444938166',
            'file': '/myemsl-dev/bundle/file.1',
            'file_storage_media': 'tape',
            'filesize': '24',
            'message': 'File was found',
            'mtime': '1444938166'
        })
        resp_bad = json.dumps({
            'bytes_per_level': '(0L, 33L, 33L, 0L, 0L)',
            'ctime': '1444938177',
            'file': '/myemsl-dev/bundle/file.2',
            'filesize': '33',
            'message': 'File was found',
            'mtime': '1444938133'
        })
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        ready = cart_utils.check_file_ready_pull(
            response,
            test_file,
            test_cart
        )
        self.assertEqual(ready, False)

        # now check for an error by sending a bad response
        ready = cart_utils.check_file_ready_pull('', test_file, test_cart)
        self.assertEqual(ready, -1)
        self.assertEqual(test_file.status, 'error')

        # now check for an error with storage media
        ready = cart_utils.check_file_ready_pull(
            resp_bad,
            test_file,
            test_cart
        )
        self.assertEqual(ready, -1)
        self.assertEqual(test_file.status, 'error')

    def test_check_file_ready_pull(self):
        """Test that checks to see if a file is ready to pull by checking the archive response."""
        response = json.dumps({
            'bytes_per_level': '(24L, 0L, 0L, 0L, 0L)',
            'ctime': '1444938166',
            'file': '/myemsl-dev/bundle/file.1',
            'file_storage_media': 'disk',
            'filesize': '24',
            'message': 'File was found',
            'mtime': '1444938166'
        })
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()
        ready = cart_utils.check_file_ready_pull(
            response,
            test_file,
            test_cart
        )
        self.assertEqual(ready['enough_space'], True)
        self.assertNotEqual(test_file.status, 'error')

    def test_delete_cart_bundle(self):
        """Test that trys to delete a cart bundle."""
        test_cart = self.create_sample_cart()
        cart_utils = Cartutils()
        os.makedirs(test_cart.bundle_path, 0o777)
        deleted = cart_utils.delete_cart_bundle(test_cart)
        self.assertEqual(deleted, True)
        self.assertEqual(test_cart.status, 'deleted')
        self.assertEqual(os.path.isdir(test_cart.bundle_path), False)

    @mock.patch.object(shutil, 'rmtree')
    def test_delete_cart_bundle_fail(self, mock_rmtree):
        """Test that trys to delete a cart bundle but fails."""
        test_cart = self.create_sample_cart()
        cart_utils = Cartutils()
        os.makedirs(test_cart.bundle_path, 0o777)
        mock_rmtree.side_effect = OSError(mock.Mock(), 'Error')
        deleted = cart_utils.delete_cart_bundle(test_cart)
        self.assertNotEqual(test_cart.status, 'deleted')
        self.assertEqual(deleted, False)
        self.assertEqual(os.path.isdir(test_cart.bundle_path), True)

    def test_set_file_status(self):
        """Test that trys to set a specific files status."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()

        cart_utils.set_file_status(
            test_file,
            test_cart,
            'error',
            'fake error'
        )
        self.assertEqual(test_file.status, 'error')
        self.assertEqual(test_file.error, 'fake error')

    def test_status_details_fail(self):
        """Test status details fail."""
        test_cart = self.create_sample_cart()
        test_file = self.create_sample_file(test_cart)
        cart_utils = Cartutils()

        # say file is way to big
        retval = cart_utils.check_status_details(
            test_cart,
            test_file,
            99999999999999999999999999999,
            1
        )
        self.assertEqual(retval, -1)

    def test_cart_no_hash_passed(self):
        """Test error with cart with no hash passed."""
        test_cart = self.create_sample_cart()
        cart_utils = Cartutils()

        data = json.loads(
            '{"fileids": [{"id":"foo.txt", "path":"1/2/3/foo.txt", "hashtype":"md5"}]}'
        )

        file_ids = data['fileids']
        retval = cart_utils.update_cart_files(test_cart, file_ids)
        self.assertNotEqual(retval, None)

    def test_lru_cart_delete(self):
        """Test that trys to delete a cart."""
        test_cart = self.create_sample_cart()
        test_cart2 = Cart.create(
            cart_uid='2',
            status='staging',
            bundle_path='/tmp/2/',
            updated_date=1
        )
        cart_utils = Cartutils()
        os.makedirs(test_cart2.bundle_path, 0o777)
        retval = cart_utils.lru_cart_delete(test_cart)
        self.assertEqual(retval, True)
        test_c2 = Cart.get(Cart.id == test_cart2.id)
        self.assertEqual(test_c2.status, 'deleted')
        # also hit error block when nothing to delete
        retval = cart_utils.lru_cart_delete(test_cart)
        self.assertEqual(retval, False)

    def test_bad_cart_status(self):
        """Test getting a status of a cart that doesnt exist."""
        cart_utils = Cartutils()
        retval = cart_utils.cart_status('2')
        self.assertEqual(retval[0], 'error')

    def test_bad_available_cart(self):
        """Test getting a cart that doesnt exist."""
        cart_utils = Cartutils()
        retval = cart_utils.available_cart('2')
        self.assertEqual(retval, None)

    @mock.patch.object(Cartutils, 'delete_cart_bundle')
    def test_bad_stage(self, mock_delete_cart):
        """Test the bad stage of a archive file."""
        test_cart = Cart.create(cart_uid='1', status='staging')

        def fake_database_connect(cls_name):  # pragma: no cover testing code
            """No error."""
            return cls_name

        def fake_database_close(cls_name):  # pragma: no cover testing code
            """No error."""
            return cls_name
        pacifica.cartd.orm.CartBase.database_close = MethodType(
            fake_database_close, pacifica.cartd.orm.CartBase)
        pacifica.cartd.orm.CartBase.database_connect = MethodType(
            fake_database_connect, pacifica.cartd.orm.CartBase)
        pacifica.cartd.orm.CartBase.throw_error = False
        mock_delete_cart.return_value = False
        cart_util = Cartutils()
        return_val = cart_util.remove_cart(test_cart.id)
        self.assertEqual(return_val, None)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Cart Object Relational Model.

Using PeeWee to implement the ORM.
"""
# disable this for classes Cart, File and Meta (within Cart and File)
# pylint: disable=too-few-public-methods
# pylint: disable=invalid-name
import datetime
import time
from peewee import PrimaryKeyField, IntegerField, CharField, DateTimeField
from peewee import ForeignKeyField, TextField, BooleanField
from peewee import Model, OperationalError
from playhouse.migrate import SchemaMigrator, migrate
from playhouse.db_url import connect
from .config import get_config

SCHEMA_MAJOR = 1
SCHEMA_MINOR = 0
DB = connect(get_config().get('database', 'peewee_url'))


class orm_sync(object):
    """
    Special module for syncing the orm.

    This module should incorporate a schema migration strategy.

    The supported versions migrating forward must be in a versions array
    containing tuples for major and minor versions.

    The version tuples are directly translated to method names in the
    orm_update class for the update between those versions.

    class orm_update:
        versions = [
            (0, 1),
            (0, 2),
            (1, 0),
            (1, 1)
        ]

        def update_0_1_to_0_2():
            ...
        def update_0_2_to_1_0():
            ...
        ...

    The body of the update should follow peewee migration practices.
    http://docs.peewee-orm.com/en/latest/peewee/playhouse.html#migrate
    """

    versions = [
        (0, 1),
        (1, 0)
    ]

    @staticmethod
    def dbconn_blocking():
        """Wait for the db connection."""
        dbcon_attempts = get_config().getint('database', 'connect_attempts')
        dbcon_wait = get_config().getint('database', 'connect_wait')
        while dbcon_attempts:
            try:
                Cart.database_connect()
                return
            except OperationalError:
                # couldnt connect, potentially wait and try again
                time.sleep(dbcon_wait)
                dbcon_attempts -= 1
        raise OperationalError('Failed database connect retry.')

    @staticmethod
    def create_tables():
        """Create the tables if they don't exist."""
        for cls in [CartSystem, Cart, File]:
            if not cls.table_exists():
                cls.create_table()
        CartSystem.get_or_create_version()

    @classmethod
    def update_0_1_to_1_0(cls):
        """Update by adding the boolean column."""
        migrator = SchemaMigrator(DB)
        migrate(
            migrator.add_column(
                'cart',
                'bundle',
                BooleanField(default=False, null=True)
            )
        )

    @classmethod
    def update_tables(cls):
        """Update the database to the current version."""
        verlist = cls.versions
        db_ver = CartSystem.get_version()
        if verlist.index(verlist[-1]) == verlist.index(db_ver):
            # we have the current version don't update
            return
        with Cart.atomic():
            for db_ver in verlist[verlist.index(db_ver):-1]:
                next_db_ver = verlist[verlist.index(db_ver)+1]
                method_name = 'update_{}_to_{}'.format(
                    '{}_{}'.format(*db_ver),
                    '{}_{}'.format(*next_db_ver)
                )
                getattr(cls, method_name)()
            CartSystem.drop_table()
            CartSystem.create_table()
            CartSystem.get_or_create_version()


class CartBase(Model):
    """Base Cart Model class."""

    @classmethod
    def atomic(cls):
        """Do the DB atomic bits."""
        # pylint: disable=no-member
        return cls._meta.database.atomic()
        # pylint: enable=no-member

    @classmethod
    def database_connect(cls):
        """
        Make sure database is connected.

        Dont reopen connection.
        """
        # pylint: disable=no-member
        if not cls._meta.database.is_closed():
            cls._meta.database.close()
        cls._meta.database.connect()
        # pylint: enable=no-member

    @classmethod
    def database_close(cls):
        """Close the database connection."""
        # pylint: disable=no-member
        if not cls._meta.database.is_closed():
            cls._meta.database.close()
        # pylint: enable=no-member

    class Meta(object):
        """Meta object containing the database connection."""

        database = DB  # This model uses the pacifica_cart database.

    def reload(self):
        """Reload my current state from the DB."""
        newer_self = self.get(self._meta.primary_key == getattr(
            self, self._meta.primary_key.name))
        for field_name in self._meta.fields.keys():
            val = getattr(newer_self, field_name)
            setattr(self, field_name, val)
        self._dirty.clear()


class CartSystem(CartBase):
    """Cart Schema Version Model."""

    part = CharField(primary_key=True)
    value = IntegerField(default=-1)

    @classmethod
    def get_or_create_version(cls):
        """Set or create the current version of the schema."""
        major = cls.get_or_create(part='major', value=SCHEMA_MAJOR)
        minor = cls.get_or_create(part='minor', value=SCHEMA_MINOR)
        return (major, minor)

    @classmethod
    def get_version(cls):
        """Get the current version as a tuple."""
        return (cls.get(part='major').value, cls.get(part='minor').value)

    @classmethod
    def is_equal(cls):
        """Check to see if schema version matches code version."""
        major, minor = cls.get_version()
        return major == SCHEMA_MAJOR and minor == SCHEMA_MINOR

    @classmethod
    def is_safe(cls):
        """Check to see if the schema version is safe for the code."""
        major, _minor = cls.get_version()
        return major == SCHEMA_MAJOR


class Cart(CartBase):
    """Cart object model."""

    id = PrimaryKeyField()
    cart_uid = CharField(default=1)
    bundle_path = CharField(default='')
    bundle = BooleanField(default=False, null=True)
    creation_date = DateTimeField(default=datetime.datetime.now)
    updated_date = DateTimeField(default=datetime.datetime.now)
    deleted_date = DateTimeField(null=True)
    status = TextField(default='waiting')
    error = TextField(default='')


class File(CartBase):
    """File object model to keep track of what's been downloaded for a cart."""

    id = PrimaryKeyField()
    cart = ForeignKeyField(Cart, to_field='id')
    file_name = CharField(default='')
    bundle_path = CharField(default='')
    hash_type = CharField(null=True)
    hash_value = CharField(null=True)
    status = TextField(default='waiting')
    error = TextField(default='')



#!/usr/bin/python
# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
"""Run the Cart Server."""
import cherrypy
from .rest import CartRoot, error_page_default
from .globals import CHERRYPY_CONFIG

cherrypy.config.update({'error_page.default': error_page_default})
# pylint doesn't realize that application is actually a callable
# pylint: disable=invalid-name
application = cherrypy.Application(CartRoot(), '/', CHERRYPY_CONFIG)
# pylint: enable=invalid-name



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Pacifica Cart Interface.

This is the main program that starts the WSGI server.
The core of the server code is in cart_interface.py.
"""
from time import sleep
from argparse import ArgumentParser, SUPPRESS
from threading import Thread
import cherrypy
from peewee import OperationalError, PeeweeException
from .orm import orm_sync, CartSystem, SCHEMA_MAJOR, SCHEMA_MINOR
from .rest import CartRoot, error_page_default
from .globals import CHERRYPY_CONFIG, CONFIG_FILE


def stop_later(doit=False):
    """Used for unit testing stop after 60 seconds."""
    if not doit:  # pragma: no cover
        return

    def sleep_then_exit():
        """
        Sleep for 60 seconds then call cherrypy exit.

        Hopefully this is long enough for the end-to-end tests to finish
        """
        sleep(90)
        cherrypy.engine.exit()
    sleep_thread = Thread(target=sleep_then_exit)
    sleep_thread.daemon = True
    sleep_thread.start()


def main():
    """Main method to start the httpd server."""
    parser = ArgumentParser(description='Run the cart server.')
    parser.add_argument('-c', '--config', metavar='CONFIG', type=str,
                        default=CONFIG_FILE, dest='config',
                        help='cart config file')
    parser.add_argument('--cpconfig', metavar='CONFIG', type=str,
                        default=CHERRYPY_CONFIG, dest='cpconfig',
                        help='cherrypy config file')
    parser.add_argument('-p', '--port', metavar='PORT', type=int,
                        default=8081, dest='port',
                        help='port to listen on')
    parser.add_argument('-a', '--address', metavar='ADDRESS',
                        default='127.0.0.1', dest='address',
                        help='address to listen on')
    parser.add_argument('--stop-after-a-moment', help=SUPPRESS,
                        default=False, dest='stop_later',
                        action='store_true')
    args = parser.parse_args()
    orm_sync.dbconn_blocking()
    if not CartSystem.is_safe():
        raise OperationalError('Database version too old {} update to {}'.format(
            '{}.{}'.format(*(CartSystem.get_version())),
            '{}.{}'.format(SCHEMA_MAJOR, SCHEMA_MINOR)
        ))
    stop_later(args.stop_later)
    cherrypy.config.update({'error_page.default': error_page_default})
    cherrypy.config.update({
        'server.socket_host': args.address,
        'server.socket_port': args.port
    })
    cherrypy.quickstart(CartRoot(), '/', args.cpconfig)


def cmd():
    """Main admin command line tool."""
    parser = ArgumentParser(description='Cart admin tool.')
    parser.add_argument(
        '-c', '--config', metavar='CONFIG', type=str, default=CONFIG_FILE,
        dest='config', help='cart config file'
    )
    subparsers = parser.add_subparsers(help='sub-command help')
    db_parser = subparsers.add_parser(
        'dbsync',
        description='Update or Create the Database.'
    )
    db_parser.set_defaults(func=dbsync)
    dbchk_parser = subparsers.add_parser(
        'dbchk',
        description='Check database against current version.'
    )
    dbchk_parser.add_argument(
        '--equal', default=False,
        dest='check_equal', action='store_true'
    )
    dbchk_parser.set_defaults(func=dbchk)
    args = parser.parse_args()
    return args.func(args)


def bool2cmdint(command_bool):
    """Convert a boolean to either 0 for true  or -1 for false."""
    if command_bool:
        return 0
    return -1


def dbchk(args):
    """Check the database for the version running."""
    orm_sync.dbconn_blocking()
    if args.check_equal:
        return bool2cmdint(CartSystem.is_equal())
    return bool2cmdint(CartSystem.is_safe())


def dbsync(_args):
    """Create or update the database."""
    orm_sync.dbconn_blocking()
    try:
        CartSystem.get_version()
    except PeeweeException:
        return orm_sync.create_tables()
    return orm_sync.update_tables()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Setup and install the cart."""
try:  # pip version 9
    from pip.req import parse_requirements
except ImportError:
    from pip._internal.req import parse_requirements
from setuptools import setup, find_packages

# parse_requirements() returns generator of pip.req.InstallRequirement objects
INSTALL_REQS = parse_requirements('requirements.txt', session='hack')

setup(
    name='pacifica-cartd',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='Pacifica Cartd',
    author='David Brown',
    author_email='dmlb2000@gmail.com',
    packages=find_packages(),
    namespace_packages=['pacifica'],
    entry_points={
        'console_scripts': [
            'pacifica-cartd=pacifica.cartd.__main__:main',
            'pacifica-cartd-cmd=pacifica.cartd.__main__:cmd'
        ]
    },
    install_requires=[str(ir.req) for ir in INSTALL_REQS]
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Configuration reading and validation module."""
from os import getenv
import logging
try:
    from ConfigParser import SafeConfigParser
except ImportError:  # pragma: no cover python 2 vs 3 issue
    from configparser import ConfigParser as SafeConfigParser
from .globals import CONFIG_FILE


def get_config():
    """Return the ConfigParser object with defaults set."""
    configparser = SafeConfigParser()
    configparser.add_section('cartd')
    configparser.set('cartd', 'volume_path', getenv(
        'VOLUME_PATH', '/tmp/'))
    configparser.set('cartd', 'lru_buffer_time', getenv(
        'LRU_BUFFER_TIME', '0'))
    configparser.set('cartd', 'bundle_task', getenv(
        'BUNDLE_TASK', 'off'))
    configparser.add_section('database')
    configparser.set('database', 'peewee_url', getenv(
        'PEEWEE_URL', 'sqliteext:///db.sqlite3'))
    configparser.set('database', 'debug_logging', getenv(
        'DATABASE_DEBUG_LOGGING', 'False'))
    configparser.set('database', 'connect_attempts', getenv(
        'DATABASE_CONNECT_ATTEMPTS', '10'))
    configparser.set('database', 'connect_wait', getenv(
        'DATABASE_CONNECT_WAIT', '20'))
    configparser.add_section('archiveinterface')
    configparser.set('archiveinterface', 'url', getenv(
        'ARCHIVE_INTERFACE_URL', 'http://127.0.0.1:8080/'))
    configparser.add_section('celery')
    configparser.read(CONFIG_FILE)
    return configparser


if get_config().getboolean('database', 'debug_logging'):  # pragma: no cover used for debugging
    LOGGER = logging.getLogger('peewee')
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(logging.StreamHandler())



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Module that contains all the amqp tasks that support the cart infrastructure."""
from __future__ import absolute_import
from os import utime, getenv
import datetime
import requests
from celery import Celery
from peewee import DoesNotExist
from .orm import Cart, File
from .utils import Cartutils
from .archive_requests import ArchiveRequests


CART_APP = Celery(
    'cart',
    broker=getenv('BROKER_URL', 'pyamqp://'),
    backend=getenv('BACKEND_URL', 'rpc://')
)


@CART_APP.task(ignore_result=True)
def create_cart(file_ids, uid, bundle):
    """Create the cart or update previous."""
    Cart.database_connect()
    mycart = Cart(cart_uid=uid, status='staging', bundle=bundle)
    mycart.save()
    stage_files.delay(file_ids, mycart.id)
    Cart.database_close()


@CART_APP.task(ignore_result=True)
def stage_files(file_ids, mycart_id):
    """Tell the files to be staged on the backend system."""
    Cart.database_connect()
    # with update or new, need to add in files
    mycart = Cart.get(Cart.id == mycart_id)
    cart_utils = Cartutils()
    file_id_error = cart_utils.update_cart_files(mycart, file_ids)
    if not file_id_error:
        get_files_locally.delay(mycart.id)
    else:
        mycart.status = 'error'
        mycart.error = 'Error parsing file Ids with error: ' + \
            str(file_id_error)
        mycart.updated_date = datetime.datetime.now()
        mycart.save()
    Cart.database_close()


@CART_APP.task(ignore_result=True)
def get_files_locally(cartid):
    """Pull the files to the local system from the backend."""
    # tell each file to be pulled
    Cart.database_connect()
    for cart_file in File.select().where(File.cart == cartid):
        stage_file_task.delay(cart_file.id)
    Cart.database_close()


@CART_APP.task(ignore_result=True)
def stage_file_task(file_id):
    """Stage the file from the archive, then call status."""
    Cart.database_connect()
    try:
        cart_file = File.get(File.id == file_id)
        mycart = cart_file.cart
        cart_utils = Cartutils()
        cart_utils.set_file_status(cart_file, mycart, 'staging', False)
        # make sure cart wasnt deleted before pulling file
        if mycart.deleted_date:
            Cart.database_close()
            return
    except DoesNotExist:
        Cart.database_close()
        return
    archive_request = ArchiveRequests()
    try:
        archive_request.stage_file(cart_file.file_name)
    except requests.exceptions.RequestException as ex:
        error_msg = 'Failed to stage with error: ' + str(ex)
        cart_utils.set_file_status(cart_file, mycart, 'error', error_msg)
        Cart.database_close()
        cart_utils.prepare_bundle(mycart.id)
        return
    # successful stage so move on to status
    Cart.database_close()
    status_file_task.delay(file_id)


@CART_APP.task(ignore_result=True)
def status_file_task(file_id):
    """Get the status from an archive file. If ready then pull the file."""
    Cart.database_connect()
    cart_file = File.get(File.id == file_id)
    mycart = cart_file.cart
    cart_utils = Cartutils()
    # make sure cart wasnt deleted before pulling file
    if mycart.deleted_date:
        Cart.database_close()
        return

    # check to see if file is available to pull from archive interface
    archive_request = ArchiveRequests()
    try:
        response = archive_request.status_file(cart_file.file_name)
    except requests.exceptions.RequestException as ex:
        error_msg = 'Failed to status file with error: ' + str(ex)
        cart_utils.set_file_status(cart_file, mycart, 'error', error_msg)
        Cart.database_close()
        cart_utils.prepare_bundle(mycart.id)
        return

    ready = cart_utils.check_file_ready_pull(response, cart_file, mycart)

    # Check to see if ready to pull.  If not recall this to check again
    # error on less then 0. No coverage on recall since it just calls the method again
    if isinstance(ready, int) and ready < 0:
        Cart.database_close()
        cart_utils.prepare_bundle(mycart.id)
        return
    elif not ready:  # pragma: no cover
        Cart.database_close()
        status_file_task.delay(file_id)
        return
    # ready so try to pull file
    pull_file.delay(file_id, ready['filepath'], ready['modtime'], False)


@CART_APP.task(ignore_result=True)
def pull_file(file_id, filepath, modtime, record_error):
    """Pull a file from the archive."""
    Cart.database_connect()
    try:
        cart_file = File.get(File.id == file_id)
        mycart = cart_file.cart
        cart_utils = Cartutils()
        # make sure cart wasnt deleted before pulling file
        if mycart.deleted_date:
            Cart.database_close()
            return
    except DoesNotExist:
        Cart.database_close()
        return

    archive_request = ArchiveRequests()
    try:
        archive_request.pull_file(
            cart_file.file_name, filepath, cart_file.hash_value, cart_file.hash_type)
        cart_utils.set_file_status(cart_file, mycart, 'staged', False)
        utime(filepath, (int(float(modtime)), int(float(modtime))))
        Cart.database_close()
    except requests.exceptions.RequestException as ex:
        # if request fails...try a second time, if that fails write error
        if record_error:
            error_msg = 'Failed to pull with error: ' + str(ex)
            cart_utils.set_file_status(cart_file, mycart, 'error', error_msg)
            Cart.database_close()
            cart_utils.prepare_bundle(mycart.id)

        else:
            pull_file.delay(file_id, filepath, modtime, True)
            Cart.database_close()

    except ValueError as ex:
        error_msg = 'Failed to pull with error: ' + str(ex)
        cart_utils.set_file_status(cart_file, mycart, 'error', error_msg)
        Cart.database_close()
        cart_utils.prepare_bundle(mycart.id)

    cart_utils.prepare_bundle(mycart.id)
